/* (C) 2003 Willem Jan Hengeveld  itsme@xs4all.nl */
#include <windows.h>
#include <Tlhelp32.h>

// !!! I am mixing up handles and processid's quite often.
//     on WinCE this is not a problem, but this needs to be
//     fixed before porting to another platform.

// from the platform builder
#define SA1100
#include "types.h"
#include "PKFUNCS.H"
#include "kernel.h"
#include "schedule.h"

#include "WINDEV.H"
#include "diskio.h"


#include "ItsUtils.h"
#include "dllversion.h"

#define PTR_DIFF(a,b)  (((BYTE*)b)-((BYTE*)a))

BOOL APIENTRY DllMain( HANDLE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH:
        case DLL_THREAD_ATTACH:
        case DLL_THREAD_DETACH:
        case DLL_PROCESS_DETACH:
            break;
    }
    return TRUE;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetVersion(
        DWORD cbInput, void *pbInput,
        DWORD *pcbOutput, void **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput=0;
    *ppbOutput=NULL;

    return ITSDLL_VERSION;
}

// returns 0 when process not found, 1 when process not killed, 2 when process killed.
ITSUTILS_API HRESULT STDAPICALLTYPE ITTerminateProcess(
        DWORD cbInput, TerminateProcessParams *pbInput,
        DWORD *pcbOutput, BYTE **ppbOutput,
        IRAPIStream *pStream)
{
    HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

    int res=0;

    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    if (Process32First(hTH, &pe))
    {
        do {

            if (wcsicmp(pbInput->wszProcessName, pe.szExeFile)==0)
            {
                res++;
                HANDLE hProc= OpenProcess(0, 0, pe.th32ProcessID);
                if (TerminateProcess(hProc, 0))
                    res++;
                CloseHandle(hProc);
                break;
            }
        } while (Process32Next(hTH, &pe));
    }

    CloseToolhelp32Snapshot(hTH);

    *ppbOutput=NULL;
    *pcbOutput=0;

    return res;
}

#define MEM_PAGE_SIZE 4096

void ReadMemory(DWORD dwStart, BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;
    DWORD addr= dwStart;

    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(MEM_PAGE_SIZE-(addr&(MEM_PAGE_SIZE-1)), dwLength-nCopied);

        if (IsBadReadPtr((void*)addr, nChunkSize))
            break;
        memcpy(&buf[nCopied], (void*)addr, nChunkSize);
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    *pdwCopied= nCopied;
}

void WriteMemory(DWORD dwStart, BYTE *buf, DWORD dwLength, DWORD *pdwCopied)
{
    DWORD nCopied= 0;
    DWORD addr= dwStart;

    while (nCopied < dwLength)
    {
        DWORD nChunkSize= min(MEM_PAGE_SIZE-(addr&(MEM_PAGE_SIZE-1)), dwLength-nCopied);

        if (IsBadWritePtr((void*)addr, nChunkSize))
            break;
        memcpy((void*)addr, &buf[nCopied], nChunkSize);
        nCopied += nChunkSize;
        addr += nChunkSize;
    }
    *pdwCopied= nCopied;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITReadProcessMemory(
        DWORD cbInput, ReadProcessMemoryParams *pbInput,
        DWORD *pcbOutput, ReadProcessMemoryResult **ppbOutput,
        IRAPIStream *pStream)
{
    ReadProcessMemoryResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut, &pOut->buffer[pbInput->nSize]);

    pOut= *ppbOutput= (ReadProcessMemoryResult*)LocalAlloc(LPTR, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD res= 0;
    if (pbInput->hProcess==NULL)
        ReadMemory(pbInput->dwOffset, &pOut->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesRead);
    else if (!ReadProcessMemory(pbInput->hProcess, (void*)pbInput->dwOffset, &pOut->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesRead))
        res= GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITWriteProcessMemory(
        DWORD cbInput, WriteProcessMemoryParams *pbInput,
        DWORD *pcbOutput, WriteProcessMemoryResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(WriteProcessMemoryResult);
    WriteProcessMemoryResult *pOut= *ppbOutput= (WriteProcessMemoryResult*)LocalAlloc(LPTR, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    if (pbInput->hProcess==NULL)
        WriteMemory(pbInput->dwOffset, &pbInput->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesWritten);
    else if (!WriteProcessMemory(pbInput->hProcess, (void*)pbInput->dwOffset, &pbInput->buffer[0], pbInput->nSize, &pOut->dwNumberOfBytesWritten))
        return GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetContext(
        DWORD cbInput, BYTE *pbInput,
        DWORD *pcbOutput, GetContextResult **ppbOutput,
        IRAPIStream *pStream)
{
    WCHAR *cmdline;

    cmdline= GetCommandLine();

    GetContextResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut, &pOut->wszCmdLine[wcslen(cmdline)+1]);

    pOut= *ppbOutput= (GetContextResult*)LocalAlloc(LPTR, *pcbOutput);

    pOut->dwProcessId= GetCurrentProcessId();
    pOut->hProcess= GetCurrentProcess();
    pOut->dwThreadId= GetCurrentThreadId();
    pOut->hThread= GetCurrentThread();
    pOut->dwCallerTrust= CeGetCallerTrust();
    pOut->dwCurrentTrust= CeGetCurrentTrust();
    wcscpy(pOut->wszCmdLine, cmdline);

    return 0;
}

int GetProcessCount(HANDLE hTH)
{
    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    int n= 0;
    if (Process32First(hTH, &pe))
    {
        do {
            n++;
        } while (Process32Next(hTH, &pe));
    }
    return n;
}

DWORD CalcHeapUsage(HANDLE hTH, DWORD dwProcessID, DWORD dwHeapID)
{
    DWORD dwTotal= 0;
    HEAPENTRY32 he;

    __try {
        he.dwSize= sizeof(HEAPENTRY32);
        if (Heap32First(hTH, &he, dwProcessID, dwHeapID))
        {
            do {
                dwTotal += he.dwBlockSize;

                he.dwSize= sizeof(HEAPENTRY32);
            } while (Heap32Next(hTH, &he));
        }
    }
    __except(1) {
        dwTotal= -1;
    }

    return dwTotal;
}

void GetProcessTimes(DWORD pid, DWORD *ptKernel, DWORD *ptUser)
{
    HDATA *hdata= (HDATA *)(0x80000000+((DWORD)pid&0x1ffffffc));

    PROCESS *pproc= (PROCESS *)hdata->pvObj;

    DWORD tKernel=0;
    DWORD tUser=0;

    THREAD *pth= pproc->pTh;
    while(pth)
    {
        tKernel += pth->dwKernTime;
        tUser   += pth->dwUserTime;

        pth= pth->pNextInProc;
    }
}

const WCHAR *GetProcessCommandLine(DWORD pid)
{
    HDATA *hdata= (HDATA *)(0x80000000+((DWORD)pid&0x1ffffffc));

    PROCESS *pproc= (PROCESS *)hdata->pvObj;

    return pproc->pcmdline;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetProcessList(
        DWORD cbInput, BYTE *pbInput,
        DWORD *pcbOutput, GetProcessListResult **ppbOutput,
        IRAPIStream *pStream)
{
    HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS|TH32CS_SNAPHEAPLIST, 0);
    if (hTH==INVALID_HANDLE_VALUE)
        return GetLastError();

    int nProcesses= GetProcessCount(hTH);
    if (nProcesses==0)
        return GetLastError();

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    GetProcessListResult *pOut=NULL;

    *pcbOutput= PTR_DIFF(pOut,  &pOut->pe[nProcesses]);

    pOut= *ppbOutput= (GetProcessListResult*)LocalAlloc(LPTR, *pcbOutput);

    pOut->nEntries= nProcesses;

    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    int i=0;
    if (Process32First(hTH, &pe))
    {
        do {
            CEPROCESSENTRY *cepe= &pOut->pe[i++];

            cepe->cntUsage= pe.cntUsage; 
            cepe->dwProcessID= pe.th32ProcessID; 
            cepe->dwDefaultHeapID= pe.th32DefaultHeapID; 
            cepe->dwModuleID= pe.th32ModuleID; 
            cepe->cntThreads= pe.cntThreads; 
            cepe->dwParentProcessID= pe.th32ParentProcessID; 
            cepe->dwFlags= pe.dwFlags; 
            cepe->dwMemoryBase= pe.th32MemoryBase;
            cepe->dwAccessKey= pe.th32AccessKey;

            wcsncpy(cepe->szExeFile, pe.szExeFile, MAX_PATH); 
            wcsncpy(cepe->szCmdLine, GetProcessCommandLine(pe.th32ProcessID), MAX_PATH);

            cepe->dwDefaultHeapSize= CalcHeapUsage(hTH, pe.th32ProcessID, pe.th32DefaultHeapID);

            GetProcessTimes(pe.th32ProcessID, &cepe->dwKernelTime, &cepe->dwUserTime);

        } while (Process32Next(hTH, &pe));
    }

    CloseToolhelp32Snapshot(hTH);

    SetProcPermissions(dwPerm);
    SetKMode(bMode);

    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITFindNextMemoryBlock(
        DWORD cbInput, FindNextMemoryBlockParams *pbInput,
        DWORD *pcbOutput, FindNextMemoryBlockResult **ppbOutput,
        IRAPIStream *pStream)
{
    int wrapped=0;

    *pcbOutput= sizeof(FindNextMemoryBlockResult);
    FindNextMemoryBlockResult *pOut= *ppbOutput= (FindNextMemoryBlockResult*)LocalAlloc(LPTR, *pcbOutput);

    DWORD addr= pbInput->dwStartSearchAddress;
    while (IsBadReadPtr((void*)addr, 4096))
    {
        if (addr+4096 < addr)
            wrapped= 1;
        addr += 4096;
    }

    pOut->dwBlockStart= addr;
    DWORD len= 0;
    while (!IsBadReadPtr((void*)addr, 4096))
    {
        if (addr+4096 < addr)
            wrapped= 1;
        addr += 4096;
        len += 4096;
    }

    pOut->dwBlockLength= len;

    return wrapped;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITGetProcessHandle(
        DWORD cbInput, GetProcessHandleParams *pbInput,
        DWORD *pcbOutput, GetProcessHandleResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput=sizeof(GetProcessHandleResult);
    *ppbOutput=(GetProcessHandleResult*)LocalAlloc(LPTR, *pcbOutput);

    (*ppbOutput)->hProc= GetProcessHandle(pbInput->wszProcessName);

    return 0;
}
HANDLE GetProcessHandle(WCHAR *wszProcessName)
{
    HANDLE hTH= CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);

    PROCESSENTRY32 pe;
    pe.dwSize= sizeof(PROCESSENTRY32);

    HANDLE hProc= INVALID_HANDLE_VALUE;
    if (Process32First(hTH, &pe))
    {
        do {

            if (wcsicmp(wszProcessName, pe.szExeFile)==0)
            {
                hProc= OpenProcess(0, 0, pe.th32ProcessID);
                if (hProc != INVALID_HANDLE_VALUE && hProc!=NULL)
                    break;
            }
        } while (Process32Next(hTH, &pe));
    }

    CloseToolhelp32Snapshot(hTH);

    return hProc;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITGetThreadTimes(
        DWORD cbInput, GetThreadTimesParams *pbInput,
        DWORD *pcbOutput, GetThreadTimesResult **ppbOutput,
        IRAPIStream *pStream)
{
    *pcbOutput= sizeof(GetThreadTimesResult);
    GetThreadTimesResult *pOut= *ppbOutput= (GetThreadTimesResult *)LocalAlloc(LPTR, *pcbOutput);

    BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    DWORD res= 0;
    if (!GetThreadTimes(pbInput->hThread, &pOut->tCreate, &pOut->tExit, &pOut->tKernel, &pOut->tUser))
        res= GetLastError();

    SetProcPermissions(dwPerm);
    SetKMode(bMode);
    return res;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITGetProcessUsageList(
        DWORD cbInput, GetProcessUsageListParams *pbInput,
        DWORD *pcbOutput, GetProcessUsageListResult **ppbOutput,
        IRAPIStream *pStream)
{
    DWORD pid= GetCurrentProcessId();

    HDATA *hi= (HDATA *)(0x80000000+((DWORD)pid&0x1ffffffc));

    HDATA *ha;

    *pcbOutput= sizeof(GetProcessUsageListResult);
    GetProcessUsageListResult *pOut= *ppbOutput= (GetProcessUsageListResult *)LocalAlloc(LPTR, *pcbOutput);

    memset(pOut, 0, *pcbOutput);

    pOut->tQuery= GetTickCount();
    ha=hi;
    do {
        if (ha->pvObj && strncmp(ha->pci->acName, "THRD", 4)==0)
        {
            THREAD *pt = (THREAD*)ha->pvObj;

            SummaryInfo &ps= pOut->list[pt->pOwnerProc->procnum];

            ps.hProc= pt->pOwnerProc->hProc;
            ps.tKernel += pt->dwKernTime;
            ps.tUser   += pt->dwUserTime;
        }

        ha= (HDATA*)ha->linkage.fwd;
    } while (ha!=hi && ha!=NULL);

    return 0;
}

//---------------------debugger----------------------------------
ITSUTILS_API HRESULT STDAPICALLTYPE ITStartDebuggingProcess(
        DWORD cbInput, StartDebuggingProcessParams *pbInput,
        DWORD *pcbOutput, StartDebuggingProcessResult **ppbOutput,
        IRAPIStream *pStream)
{
    HMODULE hDll= LoadLibrary(L"ItsUtils.dll");
    if (hDll==NULL)
        return GetLastError();
    LPTHREAD_START_ROUTINE pProc= (LPTHREAD_START_ROUTINE)GetProcAddress(hDll, L"DebugThreadProc");
    if (pProc==NULL)
        return GetLastError();

    DebuggerInfo *di= (DebuggerInfo *)LocalAlloc(LPTR, sizeof(DebuggerInfo));
    memcpy(di->magic, "ITDB", 4);
    di->evStarted= CreateEvent(NULL, false, false, NULL);   // used only once
    di->evBuffer= INVALID_HANDLE_VALUE;                     // signal new buffer entry
    di->smBuffer= INVALID_HANDLE_VALUE;                     // locks buffer info
    di->evStopped= INVALID_HANDLE_VALUE;                    // used only once
    di->dwThreadResult= 0;
    di->dwDebuggedProcessId= pbInput->dwProcessId;

    wcsncpy(di->wszExeName, pbInput->wszExeName, MAX_PATH);
    wcsncpy(di->wszCmdLine, pbInput->wszCmdLine, MAX_PATH);

    di->dwMissedEvents= 0;

    di->hDll= hDll;

    di->bufsize= 0;
    di->bufhead= 0;
    di->buftail= 0;
    di->bufmaxsize= MAX_DEBUG_EVENTS;

    di->bThreadStarted= false;
    di->bThreadTerminated= false;
    di->hThread= CreateThread(NULL, 0, pProc, di, 0, &di->dwThreadId);

    // wait until debug thread is active
    bool bOk= (WaitForSingleObject(di->evStarted, INFINITE)==WAIT_OBJECT_0);
    CloseHandle(di->evStarted);
    di->evStarted= INVALID_HANDLE_VALUE;

    *pcbOutput= sizeof(StartDebuggingProcessResult);
    StartDebuggingProcessResult *pOut= *ppbOutput= (StartDebuggingProcessResult *)LocalAlloc(LPTR, *pcbOutput);

    pOut->debugger= (DWORD)di;

    if (di->dwThreadResult)
        return di->dwThreadResult;
    else if (!di->bThreadStarted)
        return -1;
                                    
    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITStopDebuggingProcess(
        DWORD cbInput, StopDebuggingProcessParams *pbInput,
        DWORD *pcbOutput, StopDebuggingProcessResult **ppbOutput,
        IRAPIStream *pStream)
{
    DebuggerInfo *di= (DebuggerInfo *)pbInput->debugger;

    if (IsBadReadPtr(di, sizeof(DebuggerInfo)) 
            || memcmp(di->magic, "ITDB", 4))
        return -1;

    if (!di->bThreadTerminated)
    {
        // creation of event signals to thread that it must stop.
        di->evStopped= CreateEvent(NULL, true, false, NULL);     // automatic event

        // wait 5 sec until thread really stopped, before freeing up memory.
        bool bOk= (WaitForSingleObject(di->evStopped, 5000)==WAIT_OBJECT_0);
        CloseHandle(di->evStopped);
        di->evStopped= INVALID_HANDLE_VALUE;

        DWORD nExitCode;
        GetExitCodeThread(di->hThread, &nExitCode);
        TerminateThread(di->hThread, nExitCode);
    }

    CloseHandle(di->hThread);

    FreeLibrary(di->hDll);

    memset(di->magic, 0, 4);
    LocalFree(di);

    *pcbOutput= 0;
    *ppbOutput= NULL;

    return 0;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITWaitForDebugEvent(
        DWORD cbInput, WaitForDebugEventParams *pbInput,
        DWORD *pcbOutput, WaitForDebugEventResult **ppbOutput,
        IRAPIStream *pStream)
{
    DWORD res=0;

    DebuggerInfo *di= (DebuggerInfo *)pbInput->debugger;

    if (IsBadReadPtr(di, sizeof(DebuggerInfo)) 
            || memcmp(di->magic, "ITDB", 4))
        return -1;

    if (!di->bThreadStarted || di->bThreadTerminated)
        return -1;

    *pcbOutput= sizeof(WaitForDebugEventResult);
    WaitForDebugEventResult *pOut= *ppbOutput= (WaitForDebugEventResult *)LocalAlloc(LPTR, *pcbOutput);

    // either there is already something in the buffer, otherwise wait for event.
    if (!di->bufsize==0 
            || (WaitForSingleObject(di->evBuffer, INFINITE)==WAIT_OBJECT_0
                && di->bThreadStarted && !di->bThreadTerminated))
    {
        // wait until we are allowed to touch the buffer internals.
        if (WaitForSingleObject(di->smBuffer, INFINITE)==WAIT_OBJECT_0)
        {
            if (!GetDebugEvent(di, &pOut->ev))
                res= -1;

            if (!ReleaseSemaphore(di->smBuffer, 1, NULL))
                res= GetLastError();
        }
        if (!ResetEvent(di->evBuffer))
            res= GetLastError();
    }

    pOut->dwMissedEvents= di->dwMissedEvents;

    return res;
}


ITSUTILS_API DWORD DebugThreadProc(DebuggerInfo *di)
{
    DWORD res=0;
    //BOOL bMode = SetKMode(TRUE);
    DWORD dwPerm = SetProcPermissions(0xFFFFFFFF);

    // either use provided pid, or start new process
    if (di->dwDebuggedProcessId!=0)
    {
        if (!DebugActiveProcess(di->dwDebuggedProcessId))
        {
            res= di->dwThreadResult= GetLastError();
        }
    }
    else
    {
        PROCESS_INFORMATION pi;
        if (!CreateProcess(di->wszExeName, di->wszCmdLine, NULL, NULL, 0, DEBUG_PROCESS,
                NULL, NULL, NULL, &pi))
        {
            res= di->dwThreadResult= GetLastError();
        }
        else {
            di->dwDebuggedProcessId= pi.dwProcessId;
        }
    }

    di->evBuffer= CreateEvent(NULL, false, false, NULL);     // signal new buffer entry
    di->smBuffer= CreateSemaphore(NULL, 1, 1, NULL);         // locks buffer info

    // tell parent process that we are initialized and ready
    if (di->evStarted!=INVALID_HANDLE_VALUE && !SetEvent(di->evStarted))
    {
        res= di->dwThreadResult= GetLastError();
    }

    di->bThreadStarted= true;

    DEBUG_EVENT ev;

    // loop until a 'stopped' event is created
    while(!res && di->evStopped==INVALID_HANDLE_VALUE)
    {
        if (WaitForDebugEvent(&ev, 1000))
        {
            if (!ProcessDebugEvent(di, &ev))
                di->dwMissedEvents++;

            if (!ContinueDebugEvent(ev.dwProcessId, ev.dwThreadId, DBG_CONTINUE))
                res= di->dwThreadResult= GetLastError();
        }
    }
    di->bThreadTerminated= true;

    // 'stopped' event found, set it.
    if (di->evStopped!=INVALID_HANDLE_VALUE && !SetEvent(di->evStopped))
        res= GetLastError();

    // try to tell whoever is waiting for us to stop doing that.
    if (di->evBuffer!=INVALID_HANDLE_VALUE)
    {
        // wake up other thread to be able to see me.
        SetEvent(di->evBuffer);
        CloseHandle(di->evBuffer);
        di->evBuffer= INVALID_HANDLE_VALUE;
    }
    if (di->smBuffer!=INVALID_HANDLE_VALUE)
    {
        CloseHandle(di->smBuffer);
        di->smBuffer= INVALID_HANDLE_VALUE;
    }

    SetProcPermissions(dwPerm);
    //SetKMode(bMode);

    return res;
}

bool GetDebugEvent(DebuggerInfo *di, DebugEvent *ev)
{
    if (di->bufsize==0)
        return false;

    memcpy(ev, &di->buffer[di->bufhead], sizeof(DebugEvent));

    if (++di->bufhead==di->bufmaxsize)
        di->bufhead= 0;

    di->bufsize--;

    return true;
}

bool ProcessDebugEvent(DebuggerInfo *di, DEBUG_EVENT *sysev)
{
    if (di->bufsize==di->bufmaxsize)
        return false;
    if (WaitForSingleObject(di->smBuffer, INFINITE)==WAIT_OBJECT_0)
    {
        if (di->bufsize==di->bufmaxsize)
            return false;

        DebugEvent *ev= &di->buffer[di->buftail];

        ev->dwProcessId= sysev->dwProcessId;
        ev->dwThreadId= sysev->dwThreadId;
        ev->dwDebugEventCode= sysev->dwDebugEventCode;

        ParseDebugEvent(ev->description, sizeof(ev->description), sysev);

        if (++di->buftail==di->bufmaxsize)
            di->buftail= 0;

        ++di->bufsize;

        if (!ReleaseSemaphore(di->smBuffer, 1, NULL))
            return false;
        if (!SetEvent(di->evBuffer))
            return false;
    }

    return true;
}

//--------------converting a debug event to a string
int ParseDebugException(char *buf, int bufsize, DWORD dwProcessId, EXCEPTION_DEBUG_INFO *exception);
int ParseDebugCreateThread(char *buf, int bufsize, DWORD dwProcessId, CREATE_THREAD_DEBUG_INFO *createThread);
int ParseDebugCreateProcessInfo(char *buf, int bufsize, DWORD dwProcessId, CREATE_PROCESS_DEBUG_INFO *createProcessInfo);
int ParseDebugExitThread(char *buf, int bufsize, DWORD dwProcessId, EXIT_THREAD_DEBUG_INFO *exitThread);
int ParseDebugExitProcess(char *buf, int bufsize, DWORD dwProcessId, EXIT_PROCESS_DEBUG_INFO *exitProcess);
int ParseDebugLoadDll(char *buf, int bufsize, DWORD dwProcessId, LOAD_DLL_DEBUG_INFO *loadDll);
int ParseDebugUnloadDll(char *buf, int bufsize, DWORD dwProcessId, UNLOAD_DLL_DEBUG_INFO *unloadDll);
int ParseDebugString(char *buf, int bufsize, DWORD dwProcessId, OUTPUT_DEBUG_STRING_INFO *debugString);
int ParseDebugRipInfo(char *buf, int bufsize, DWORD dwProcessId, RIP_INFO *ripInfo);


int ParseDebugEvent(char *buf, int bufsize, DEBUG_EVENT *event)
{
    int n=0;
    switch(event->dwDebugEventCode)
    {
    case EXCEPTION_DEBUG_EVENT:     n+=ParseDebugException(buf+n, bufsize-n, event->dwProcessId, &event->u.Exception); break;
    case CREATE_THREAD_DEBUG_EVENT: n+=ParseDebugCreateThread(buf+n, bufsize-n, event->dwProcessId, &event->u.CreateThread); break;
    case CREATE_PROCESS_DEBUG_EVENT:n+=ParseDebugCreateProcessInfo(buf+n, bufsize-n, event->dwProcessId, &event->u.CreateProcessInfo); break;
    case EXIT_THREAD_DEBUG_EVENT:   n+=ParseDebugExitThread(buf+n, bufsize-n, event->dwProcessId, &event->u.ExitThread); break;
    case EXIT_PROCESS_DEBUG_EVENT:  n+=ParseDebugExitProcess(buf+n, bufsize-n, event->dwProcessId, &event->u.ExitProcess); break;
    case LOAD_DLL_DEBUG_EVENT:      n+=ParseDebugLoadDll(buf+n, bufsize-n, event->dwProcessId, &event->u.LoadDll); break;
    case UNLOAD_DLL_DEBUG_EVENT:    n+=ParseDebugUnloadDll(buf+n, bufsize-n, event->dwProcessId, &event->u.UnloadDll); break;
    case OUTPUT_DEBUG_STRING_EVENT: n+=ParseDebugString(buf+n, bufsize-n, event->dwProcessId, &event->u.DebugString); break;
    case RIP_EVENT:                 n+=ParseDebugRipInfo(buf+n, bufsize-n, event->dwProcessId, &event->u.RipInfo); break;
    default:
        n+= _snprintf(buf+n, bufsize-n, "unknown debug event pid=%08lx tid=%08lx ev=%08lx", event->dwProcessId, event->dwThreadId, event->dwDebugEventCode);
    }
    return n;
}

// provide some info in case other things fail.
TCHAR *StrInfo(DWORD dwProcessId, void *ptr, int len, int uni)
{
    static TCHAR tbuf[256];
    _sntprintf(tbuf, 256, L"[%08lx:%08lx L%04x uni=%d]", dwProcessId, ptr, len, uni);
    return tbuf;
}
TCHAR *ReadProcessWString(DWORD dwProcessId, void *ptr, int len)
{
    static TCHAR buf[1024];
    DWORD nCopied;
    if (!ReadProcessMemory((HANDLE)dwProcessId, ptr, buf, min(len, sizeof(buf)-1), &nCopied))
        return 0;
    buf[nCopied]= 0;

    return buf;
}
char *ReadProcessString(DWORD dwProcessId, void *ptr, int len)
{
    static char buf[1024];
    DWORD nCopied;
    if (!ReadProcessMemory((HANDLE)dwProcessId, ptr, buf, min(len, sizeof(buf)-1), &nCopied))
        return 0;
    buf[nCopied]= 0;

    return buf;
}
DWORD ReadProcessDword(DWORD dwProcessId, void *ptr)
{
    DWORD dw;
    DWORD nCopied;
    if (!ReadProcessMemory((HANDLE)dwProcessId, ptr, &dw, sizeof(DWORD), &nCopied))
        return 0;

    return dw;
}

TCHAR *PtrToString(DWORD dwProcessId, void *ptr, int len, int uni)
{
    if (ptr==NULL)
        return L"(NULL)";
    else if (uni)
    {
        TCHAR *tbuf= ReadProcessWString(dwProcessId, ptr, len);
        if (tbuf==NULL)
            return StrInfo(dwProcessId, ptr, len, uni);

        return tbuf;
    }
    else {
        static TCHAR tbuf[256];
        char *buf= ReadProcessString(dwProcessId, ptr, len);

        if (buf==NULL)
            return StrInfo(dwProcessId, ptr, len, uni);
        _sntprintf(tbuf, 256, L"%hs", buf);
        return tbuf;
    }
}
TCHAR *PtrptrToString(DWORD dwProcessId, void *pptr, int len, int uni)
{
    if (pptr==NULL)
        return L"((NULL))";

    void *ptr= (void*)ReadProcessDword(dwProcessId, pptr);

    return PtrToString(dwProcessId, ptr, len, uni);
}

int ParseDebugException(char *buf, int bufsize, DWORD dwProcessId, EXCEPTION_DEBUG_INFO *exception)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "debugException first=%d", exception->dwFirstChance);
    n+= _snprintf(buf+n, bufsize-n, " code=%d flags=%08x next=%08x addr=%08x npar=%d", 
            exception->ExceptionRecord.ExceptionCode,
            exception->ExceptionRecord.ExceptionFlags,
            exception->ExceptionRecord.ExceptionRecord,
            exception->ExceptionRecord.ExceptionAddress,
            exception->ExceptionRecord.NumberParameters);
    for(DWORD i=0 ; i<exception->ExceptionRecord.NumberParameters ; i++)
    {
        n+= _snprintf(buf+n, bufsize-n, " %d", exception->ExceptionRecord.ExceptionInformation[i]);
    }
    return n;
}

int ParseDebugCreateThread(char *buf, int bufsize, DWORD dwProcessId, CREATE_THREAD_DEBUG_INFO *createThread)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "createThread(hthread=%08x base=%08x start=%08x)", 
        createThread->hThread,
        createThread->lpThreadLocalBase,
        createThread->lpStartAddress);
    return n;
}

int ParseDebugCreateProcessInfo(char *buf, int bufsize, DWORD dwProcessId, CREATE_PROCESS_DEBUG_INFO *createProcessInfo)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "createProcess(file=%d hproc=%08x hthread=%08x image=%08x debug=%08x(%08x)", 
            createProcessInfo->hFile,
            createProcessInfo->hProcess,
            createProcessInfo->hThread,
            createProcessInfo->lpBaseOfImage,
            createProcessInfo->dwDebugInfoFileOffset,
            createProcessInfo->nDebugInfoSize);
    n+= _snprintf(buf+n, bufsize-n, "       base=%08x start=%08x name='%ls'", 
            createProcessInfo->lpThreadLocalBase,
            createProcessInfo->lpStartAddress,
            PtrToString(
                    dwProcessId,
                    createProcessInfo->lpImageName,
                    128,        // todo ..... what should this be ??
                    createProcessInfo->fUnicode));
    return n;
}

int ParseDebugExitThread(char *buf, int bufsize, DWORD dwProcessId, EXIT_THREAD_DEBUG_INFO *exitThread)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "exitThread code=%d", exitThread->dwExitCode);
    return n;
}

int ParseDebugExitProcess(char *buf, int bufsize, DWORD dwProcessId, EXIT_PROCESS_DEBUG_INFO *exitProcess)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "exitProcess code=%d", exitProcess->dwExitCode);
    return n;
}

int ParseDebugLoadDll(char *buf, int bufsize, DWORD dwProcessId, LOAD_DLL_DEBUG_INFO *loadDll)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "loadDll file=%d base=%08x debug=%08x(%08x) name='%ls'", 
            loadDll->hFile,
            loadDll->lpBaseOfDll,
            loadDll->dwDebugInfoFileOffset,
            loadDll->nDebugInfoSize,
            PtrToString(
                    dwProcessId,
                    loadDll->lpImageName,
                    128,        // todo ..... what should this be ??
                    loadDll->fUnicode));
    return n;
}

int ParseDebugUnloadDll(char *buf, int bufsize, DWORD dwProcessId, UNLOAD_DLL_DEBUG_INFO *unloadDll)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "unloadDll base=%08x", unloadDll->lpBaseOfDll);
    return n;
}

int ParseDebugString(char *buf, int bufsize, DWORD dwProcessId, OUTPUT_DEBUG_STRING_INFO *debugString)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "debugString '%ls'", 
            PtrToString(dwProcessId,
                    debugString->lpDebugStringData,
                    debugString->nDebugStringLength,
                    debugString->fUnicode));
    return n;
}

int ParseDebugRipInfo(char *buf, int bufsize, DWORD dwProcessId, RIP_INFO *ripInfo)
{
    int n=0;
    n+= _snprintf(buf+n, bufsize-n, "ripinfo error=%d type=%d", ripInfo->dwError, ripInfo->dwType);
    return n;
}

//--------------------------------
int GetDiskInfo(HANDLE hDsk, DISK_INFO &info)
{
    DWORD nReturned;

    // various ce versions use all possible combinations, so we try 'm all.
    if (DeviceIoControl(hDsk, IOCTL_DISK_GETINFO, &info, sizeof(info), NULL, 0, &nReturned, NULL))
        return 1;
    if (DeviceIoControl(hDsk, IOCTL_DISK_GETINFO, NULL, 0, &info, sizeof(info), &nReturned, NULL))
        return 2;
    if (DeviceIoControl(hDsk, DISK_IOCTL_GETINFO, &info, sizeof(info), NULL, 0, &nReturned, NULL))
        return 3;
    if (DeviceIoControl(hDsk, DISK_IOCTL_GETINFO, NULL, 0, &info, sizeof(info), &nReturned, NULL))
        return 4;

    return 0;
}
HANDLE OpenDisk(int n)
{
    TCHAR diskname[16];
    _sntprintf(diskname, 16, L"DSK%u:", n);

    HANDLE hDsk= CreateFile( diskname,
            GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
            NULL, OPEN_EXISTING, 0, 0 );
    if (hDsk==NULL || hDsk==INVALID_HANDLE_VALUE)
    {
        return NULL;
    }
    return hDsk;
}

/*
    TCHAR ss[256]; _sntprintf(ss, 256, L"req: %d %08lx %08lx", pbInput->dwDiskNr, pbInput->dwOffset, pbInput->dwSize);
    MessageBox(0, ss, L"req", 0);
    _sntprintf(ss, 256, L"reply: %d %08lx %08lx -> %d, %d, %08lx %08lx", 
            pbInput->dwDiskNr, pbInput->dwOffset, pbInput->dwSize,
            out->dwNumberOfBytesRead, out->start, out->number, outsize);
    MessageBox(0, ss, L"reply", 0);
 */
bool DiskReadBlocks(HANDLE hDsk, int nStart, int nBlocks, const DISK_INFO& info,
                BYTE* buffer, DWORD* pnBytesRead)
{
    SG_REQ req;

    req.sr_start= nStart;
    req.sr_num_sec= nBlocks;
    req.sr_num_sg= 1;
    req.sr_status= 0;
    req.sr_callback= NULL;
    req.sr_sglist[0].sb_len= nBlocks*info.di_bytes_per_sect;
    req.sr_sglist[0].sb_buf= buffer;

    DWORD nReturned=0;
    if (!DeviceIoControl(hDsk, DISK_IOCTL_READ, &req, sizeof(req), NULL, NULL, &nReturned, NULL))
        return false;

    *pnBytesRead= nBlocks*info.di_bytes_per_sect;

    return true;
}
bool DiskWriteBlocks(HANDLE hDsk, int nStart, int nBlocks, const DISK_INFO& info,
            const BYTE* buffer, DWORD* pnBytesWritten)
{
    SG_REQ req;

    req.sr_start= nStart;
    req.sr_num_sec= nBlocks;
    req.sr_num_sg= 1;
    req.sr_status= 0;
    req.sr_callback= NULL;
    req.sr_sglist[0].sb_len= nBlocks*info.di_bytes_per_sect;
    req.sr_sglist[0].sb_buf= (BYTE*)buffer;

    DWORD nReturned=0;
    if (!DeviceIoControl(hDsk, DISK_IOCTL_WRITE, &req, sizeof(req), NULL, NULL, &nReturned, NULL))
        return false;

    *pnBytesWritten= nBlocks*info.di_bytes_per_sect;

    return true;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITSDCardInfo(
        DWORD cbInput, SDCardInfoParams *pbInput,
        DWORD *pcbOutput, SDCardInfoResult **ppbOutput,
        IRAPIStream *pStream)
{
    HRESULT res= 0;

    *ppbOutput= NULL;
    *pcbOutput= 0;

    HANDLE hDsk= OpenDisk(pbInput->dwDiskNr);
    if (hDsk==NULL)
        return GetLastError();
    DISK_INFO info;
    if (!GetDiskInfo(hDsk, info))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        return res;
    }
    DWORD outsize= sizeof(SDCardInfoResult);
    SDCardInfoResult *out= (SDCardInfoResult*)LocalAlloc(LPTR, outsize);

    out->blockSize= info.di_bytes_per_sect;
    out->totalBlocks= info.di_total_sectors;

    CloseHandle(hDsk);
    *ppbOutput= out;
    *pcbOutput= outsize;

    return 0;
}

ITSUTILS_API HRESULT STDAPICALLTYPE ITReadSDCard(
        DWORD cbInput, ReadSDCardParams *pbInput,
        DWORD *pcbOutput, ReadSDCardResult **ppbOutput,
        IRAPIStream *pStream)
{
    HRESULT res= 0;

    *ppbOutput= NULL;
    *pcbOutput= 0;

    HANDLE hDsk= OpenDisk(pbInput->dwDiskNr);
    if (hDsk==NULL)
        return GetLastError();
    DISK_INFO info;
    if (!GetDiskInfo(hDsk, info))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        return res;
    }
    DWORD outsize= sizeof(ReadSDCardResult)+pbInput->dwSize;
    ReadSDCardResult *out= (ReadSDCardResult *)LocalAlloc(LPTR, outsize);

    if (!DiskReadBlocks(hDsk, pbInput->dwOffset / info.di_bytes_per_sect, pbInput->dwSize / info.di_bytes_per_sect, info, out->buffer, &out->dwNumberOfBytesRead))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        LocalFree(out);
        return res;
    }

    out->start= pbInput->dwOffset / info.di_bytes_per_sect;
    out->number= pbInput->dwSize / info.di_bytes_per_sect;

    CloseHandle(hDsk);
    *ppbOutput= out;
    *pcbOutput= outsize;

    return 0;
}


ITSUTILS_API HRESULT STDAPICALLTYPE ITWriteSDCard(
        DWORD cbInput, WriteSDCardParams *pbInput,
        DWORD *pcbOutput, WriteSDCardResult **ppbOutput,
        IRAPIStream *pStream)
{
    HRESULT res= 0;

    *ppbOutput= NULL;
    *pcbOutput= 0;

    HANDLE hDsk= OpenDisk(pbInput->dwDiskNr);
    if (hDsk==NULL)
        return GetLastError();
    DISK_INFO info;
    if (!GetDiskInfo(hDsk, info))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        return res;
    }
    WriteSDCardResult *out= (WriteSDCardResult *)LocalAlloc(LPTR, sizeof(WriteSDCardResult));

    if (!DiskWriteBlocks(hDsk, pbInput->dwOffset / info.di_bytes_per_sect, pbInput->dwSize / info.di_bytes_per_sect, info, pbInput->buffer, &out->dwNumberOfBytesWritten))
    {
        res= GetLastError();
        CloseHandle(hDsk);
        LocalFree(out);
        return res;
    }
    CloseHandle(hDsk);
    *ppbOutput= out;
    *pcbOutput= sizeof(WriteSDCardResult);

    return 0;
}

//--------------------------------------------------------

#define IOCTL_HAL_OAK_START                 5000
#define IOCTL_HAL_XIP_BASE                  IOCTL_HAL_OAK_START + 500

/*++

	IOCTL_HAL_WRITE_XIP

	Writes an XIP region to flash
	
	Input Buffer is a WRITE_XIP_HEADER followed immediately by dwNumRecords
    of XIP_RECORD_INFO structures

	Output buffer is unused.

--*/


#define IOCTL_HAL_WRITE_XIP_3   CTL_CODE(FILE_DEVICE_HAL, IOCTL_HAL_XIP_BASE + 1, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_HAL_WRITE_XIP_4   CTL_CODE(FILE_DEVICE_HAL, 0x2e, METHOD_BUFFERED, FILE_ANY_ACCESS)

        
typedef struct _WRITE_XIP_HEADER
{
	LPVOID  pvPhysAddr;     // physical starting address of XIP
    DWORD	dwLength;       // length of XIP
    DWORD   dwNumRecords;   // number of records in XIP

} WRITE_XIP_HEADER, *PWRITE_XIP_HEADER;

typedef struct _XIP_RECORD_INFO {

    LPVOID  pvPhysStart;    // physical starting address of record, dest addr
    DWORD   dwPhysLen;      // physical len, in bytes of record
    DWORD   dwCheckSum;     // sum of all the bytes in the record data
    VOID UNALIGNED * pvRAMStart;     // virtual address of record in RAM, source addr

}   XIP_RECORD_INFO, *PXIP_RECORD_INFO;

ITSUTILS_API HRESULT STDAPICALLTYPE ITFlashROM(
		DWORD cbInput, FlashROMParams *pbInput,
		DWORD *pcbOutput, FlashROMResult **ppbOutput,
		IRAPIStream *pStream)
{
    *pcbOutput= sizeof(FlashROMResult);
    FlashROMResult *pOut= *ppbOutput= (FlashROMResult*)LocalAlloc(LPTR, *pcbOutput);
    struct {
        WRITE_XIP_HEADER hdr;
        XIP_RECORD_INFO info;
    } xipcmd;

    xipcmd.hdr.pvPhysAddr= (void*)pbInput->dwOffset;
    xipcmd.hdr.dwLength= pbInput->dwSize;
    xipcmd.hdr.dwNumRecords= 1;
    xipcmd.info.pvPhysStart= (void*)pbInput->dwOffset;
    xipcmd.info.dwPhysLen= pbInput->dwSize;
    xipcmd.info.dwCheckSum= pbInput->dwCheckSum;
    xipcmd.info.pvRAMStart= (VOID*)pbInput->buffer;

    OSVERSIONINFO version;
    if (!GetVersionEx(&version))
        return GetLastError();

    DWORD ioctlcode;
    if (version.dwMajorVersion==3)
        ioctlcode=IOCTL_HAL_WRITE_XIP_3;
    else
        ioctlcode=IOCTL_HAL_WRITE_XIP_4;

    if (!KernelIoControl(ioctlcode, &xipcmd, sizeof(xipcmd), &pOut->dwResult, sizeof(DWORD), &pOut->dwReturned))
        return GetLastError();

    return 0;
}

